@@ -26,12 +26,13 @@
 #include <string.h>
 #ifdef HAVE_FFMPEG_UNINSTALLED
 #include <avformat.h>
+#ifdef HAVE_AVI_H
 #include <avi.h>
+#endif
 #else
 #include <ffmpeg/avformat.h>
 #include <ffmpeg/avi.h>
 #endif
-
 #include <gst/gst.h>
 
 #include "gstffmpeg.h"
@@ -50,6 +51,7 @@ struct _GstFFStream
   GstClockTime last_ts;
   gboolean discont;
   gboolean eos;
+  GstFlowReturn last_flow;
 };
 
 struct _GstFFMpegDemux
@@ -99,9 +101,7 @@ struct _GstFFMpegDemuxClass
   GstPadTemplate *audiosrctempl;
 };
 
-static GHashTable *global_plugins;
-
-/* A number of functon prototypes are given so we can refer to them later. */
+/* A number of function prototypes are given so we can refer to them later. */
 static void gst_ffmpegdemux_class_init (GstFFMpegDemuxClass * klass);
 static void gst_ffmpegdemux_base_init (GstFFMpegDemuxClass * klass);
 static void gst_ffmpegdemux_init (GstFFMpegDemux * demux);
@@ -124,6 +124,8 @@ gst_ffmpegdemux_send_event (GstElement * element, GstEvent * event);
 static GstStateChangeReturn
 gst_ffmpegdemux_change_state (GstElement * element, GstStateChange transition);
 
+#define GST_FFDEMUX_PARAMS_QDATA g_quark_from_static_string("ffdemux-params")
+
 static GstElementClass *parent_class = NULL;
 
 static const gchar *
@@ -164,17 +166,14 @@ gst_ffmpegdemux_averror (gint av_errno)
 static void
 gst_ffmpegdemux_base_init (GstFFMpegDemuxClass * klass)
 {
-  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
   GstFFMpegDemuxClassParams *params;
   GstElementDetails details;
   GstPadTemplate *sinktempl, *audiosrctempl, *videosrctempl;
 
-  params = g_hash_table_lookup (global_plugins,
-      GINT_TO_POINTER (G_OBJECT_CLASS_TYPE (gobject_class)));
-  if (!params)
-    params = g_hash_table_lookup (global_plugins, GINT_TO_POINTER (0));
-  g_assert (params);
+  params = (GstFFMpegDemuxClassParams *)
+      g_type_get_qdata (G_OBJECT_CLASS_TYPE (klass), GST_FFDEMUX_PARAMS_QDATA);
+  g_assert (params != NULL);
 
   /* construct the element details struct */
   details.longname = g_strdup_printf ("FFMPEG %s demuxer",
@@ -508,9 +507,15 @@ gst_ffmpegdemux_perform_seek (GstFFMpegDemux * demux, GstEvent * event)
 
   /* and prepare to continue streaming */
   if (flush) {
+    gint n;
+
     /* send flush stop, peer will accept data and events again. We
      * are not yet providing data as we still have the STREAM_LOCK. */
     gst_ffmpegdemux_push_event (demux, gst_event_new_flush_stop ());
+    for (n = 0; n < MAX_STREAMS; ++n) {
+      if (demux->streams[n])
+        demux->streams[n]->last_flow = GST_FLOW_OK;
+    }
   } else if (res && demux->running) {
     /* we are running the current segment and doing a non-flushing seek,
      * close the segment first based on the last_stop. */
@@ -585,14 +590,20 @@ gst_ffmpegdemux_src_event (GstPad * pad, GstEvent * event)
   switch (GST_EVENT_TYPE (event)) {
     case GST_EVENT_SEEK:
       res = gst_ffmpegdemux_perform_seek (demux, event);
+      gst_event_unref (event);
       break;
+    case GST_EVENT_LATENCY:
+      res = gst_pad_push_event (demux->sinkpad, event);
+      break;
+    case GST_EVENT_NAVIGATION:
+    case GST_EVENT_QOS:
     default:
       res = FALSE;
+      gst_event_unref (event);
       break;
   }
 
   gst_object_unref (demux);
-  gst_event_unref (event);
 
   return res;
 }
@@ -698,8 +709,12 @@ gst_ffmpegdemux_src_query (GstPad * pad, GstQuery * query)
 
       timeduration =
           gst_ffmpeg_time_ff_to_gst (avstream->duration, avstream->time_base);
-      if (!(GST_CLOCK_TIME_IS_VALID (timeduration)))
-        break;
+      if (!(GST_CLOCK_TIME_IS_VALID (timeduration))) {
+        /* use duration of complete file if the stream duration is not known */
+        timeduration = demux->duration;
+        if (!(GST_CLOCK_TIME_IS_VALID (timeduration)))
+          break;
+      }
 
       switch (format) {
         case GST_FORMAT_TIME:
@@ -783,6 +798,34 @@ gst_ffmpegdemux_src_convert (GstPad * pad,
 }
 #endif
 
+static GstFlowReturn
+gst_ffmpegdemux_aggregated_flow (GstFFMpegDemux * demux)
+{
+  gint n;
+  GstFlowReturn res = GST_FLOW_OK;
+  gboolean have_ok = FALSE;
+
+  for (n = 0; n < MAX_STREAMS; n++) {
+    GstFFStream *s = demux->streams[n];
+
+    if (s) {
+      res = MIN (res, s->last_flow);
+
+      if (GST_FLOW_IS_SUCCESS (s->last_flow))
+        have_ok = TRUE;
+    }
+  }
+
+  /* NOT_LINKED is OK, if at least one pad is linked */
+  if (res == GST_FLOW_NOT_LINKED && have_ok)
+    res = GST_FLOW_OK;
+
+  GST_DEBUG_OBJECT (demux, "Returning aggregated value of %s",
+      gst_flow_get_name (res));
+
+  return res;
+}
+
 static GstFFStream *
 gst_ffmpegdemux_get_stream (GstFFMpegDemux * demux, AVStream * avstream)
 {
@@ -812,6 +855,7 @@ gst_ffmpegdemux_get_stream (GstFFMpegDemux * demux, AVStream * avstream)
   stream->discont = TRUE;
   stream->avstream = avstream;
   stream->last_ts = GST_CLOCK_TIME_NONE;
+  stream->last_flow = GST_FLOW_OK;
 
   switch (ctx->codec_type) {
     case CODEC_TYPE_VIDEO:
@@ -997,7 +1041,7 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
   if (res < 0)
     goto open_failed;
 
-  res = av_find_stream_info (demux->context);
+  res = gst_ffmpeg_av_find_stream_info (demux->context);
   GST_DEBUG_OBJECT (demux, "av_find_stream_info returned %d", res);
   if (res < 0)
     goto no_info;
@@ -1078,18 +1122,31 @@ gst_ffmpegdemux_type_find (GstTypeFind * tf, gpointer priv)
   GstFFMpegDemuxClassParams *params = (GstFFMpegDemuxClassParams *) priv;
   AVInputFormat *in_plugin = params->in_plugin;
   gint res = 0;
+  guint64 length;
+
+  /* We want GST_FFMPEG_TYPE_FIND_SIZE bytes, but if the file is shorter than
+   * that we'll give it a try... */
+  length = gst_type_find_get_length (tf);
+  if (length == 0 || length > GST_FFMPEG_TYPE_FIND_SIZE)
+    length = GST_FFMPEG_TYPE_FIND_SIZE;
 
   if (in_plugin->read_probe &&
-      (data = gst_type_find_peek (tf, 0, GST_FFMPEG_TYPE_FIND_SIZE)) != NULL) {
+      (data = gst_type_find_peek (tf, 0, length)) != NULL) {
     AVProbeData probe_data;
 
     probe_data.filename = "";
     probe_data.buf = data;
-    probe_data.buf_size = GST_FFMPEG_TYPE_FIND_SIZE;
+    probe_data.buf_size = length;
 
     res = in_plugin->read_probe (&probe_data);
     if (res > 0) {
       res = MAX (1, res * GST_TYPE_FIND_MAXIMUM / AVPROBE_SCORE_MAX);
+      /* Restrict the probability for MPEG-TS streams, because there is
+       * probably a better version in plugins-base, if the user has a recent
+       * plugins-base */
+      if (!strcmp (in_plugin->name, "mpegts"))
+        res = MIN (res, GST_TYPE_FIND_POSSIBLE);
+
       gst_type_find_suggest (tf, res, params->sinkcaps);
     }
   }
@@ -1142,6 +1199,11 @@ gst_ffmpegdemux_loop (GstPad * pad)
     stream->last_ts = timestamp;
   }
   duration = gst_ffmpeg_time_ff_to_gst (pkt.duration, avstream->time_base);
+  if (G_UNLIKELY (!duration)) {
+    GST_WARNING_OBJECT (demux, "invalid buffer duration, setting to NONE");
+    duration = GST_CLOCK_TIME_NONE;
+  }
+
 
   GST_DEBUG_OBJECT (demux,
       "pkt pts:%" GST_TIME_FORMAT
@@ -1193,7 +1255,15 @@ gst_ffmpegdemux_loop (GstPad * pad)
       "Sending out buffer time:%" GST_TIME_FORMAT " size:%d",
       GST_TIME_ARGS (timestamp), GST_BUFFER_SIZE (outbuf));
 
-  ret = gst_pad_push (srcpad, outbuf);
+  ret = stream->last_flow = gst_pad_push (srcpad, outbuf);
+
+  /* if a pad is in e.g. WRONG_STATE, we want to pause to unlock the STREAM_LOCK */
+  if ((ret != GST_FLOW_OK)
+      && ((ret = gst_ffmpegdemux_aggregated_flow (demux)) != GST_FLOW_OK)) {
+    GST_WARNING_OBJECT (demux, "stream_movi flow: %s / %s",
+        gst_flow_get_name (stream->last_flow), gst_flow_get_name (ret));
+    goto pause;
+  }
 
 done:
   /* can destroy the packet now */
@@ -1388,8 +1458,6 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
 
   in_plugin = first_iformat;
 
-  global_plugins = g_hash_table_new (NULL, NULL);
-
   while (in_plugin) {
     gchar *type_name, *typefind_name;
     gchar *p, *name = NULL;
@@ -1428,42 +1496,42 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         !strcmp (in_plugin->name, "wav") ||
         !strcmp (in_plugin->name, "au") ||
         !strcmp (in_plugin->name, "tta") ||
-	!strcmp (in_plugin->name, "rm") ||
-	!strcmp (in_plugin->name, "amr"))
+        !strcmp (in_plugin->name, "rm") || !strcmp (in_plugin->name, "amr"))
       register_typefind_func = FALSE;
 
     /* Set the rank of demuxers know to work to MARGINAL.
      * Set demuxers for which we already have another implementation to NONE
      * Set All others to NONE*/
     if (!strcmp (in_plugin->name, "flv") ||
-	!strcmp (in_plugin->name, "wsvqa") ||
-	!strcmp (in_plugin->name, "wsaud") ||
-	!strcmp (in_plugin->name, "wc3movie") ||
-	!strcmp (in_plugin->name, "voc") ||
-	!strcmp (in_plugin->name, "tta") ||
-	!strcmp (in_plugin->name, "swf") ||
-	!strcmp (in_plugin->name, "sol") ||
-	!strcmp (in_plugin->name, "smk") ||
-	!strcmp (in_plugin->name, "vmd") ||
-	!strcmp (in_plugin->name, "film_cpk") ||
-	!strcmp (in_plugin->name, "ingenient") ||
-	!strcmp (in_plugin->name, "psxstr") ||
-	!strcmp (in_plugin->name, "nuv") ||
-	!strcmp (in_plugin->name, "nut") ||
-	!strcmp (in_plugin->name, "nsv") ||
-	!strcmp (in_plugin->name, "mxf") ||
-	!strcmp (in_plugin->name, "mmf") ||
-	!strcmp (in_plugin->name, "mm") ||
-	!strcmp (in_plugin->name, "ipmovie") ||
-	!strcmp (in_plugin->name, "RoQ") ||
-	!strcmp (in_plugin->name, "idcin") ||
-	!strcmp (in_plugin->name, "gxf") ||
-	!strcmp (in_plugin->name, "ffm") ||
-	!strcmp (in_plugin->name, "ea") ||
-	!strcmp (in_plugin->name, "daud") ||
-	!strcmp (in_plugin->name, "avs") ||
-	!strcmp (in_plugin->name, "aiff") ||
-	!strcmp (in_plugin->name, "4xm"))
+        !strcmp (in_plugin->name, "wsvqa") ||
+        !strcmp (in_plugin->name, "wsaud") ||
+        !strcmp (in_plugin->name, "wc3movie") ||
+        !strcmp (in_plugin->name, "voc") ||
+        !strcmp (in_plugin->name, "tta") ||
+        !strcmp (in_plugin->name, "swf") ||
+        !strcmp (in_plugin->name, "sol") ||
+        !strcmp (in_plugin->name, "smk") ||
+        !strcmp (in_plugin->name, "vmd") ||
+        !strcmp (in_plugin->name, "film_cpk") ||
+        !strcmp (in_plugin->name, "ingenient") ||
+        !strcmp (in_plugin->name, "psxstr") ||
+        !strcmp (in_plugin->name, "nuv") ||
+        !strcmp (in_plugin->name, "nut") ||
+        !strcmp (in_plugin->name, "nsv") ||
+        !strcmp (in_plugin->name, "mxf") ||
+        !strcmp (in_plugin->name, "mmf") ||
+        !strcmp (in_plugin->name, "mm") ||
+        !strcmp (in_plugin->name, "ipmovie") ||
+        !strcmp (in_plugin->name, "RoQ") ||
+        !strcmp (in_plugin->name, "idcin") ||
+        !strcmp (in_plugin->name, "gxf") ||
+        !strcmp (in_plugin->name, "ffm") ||
+        !strcmp (in_plugin->name, "ea") ||
+        !strcmp (in_plugin->name, "daud") ||
+        !strcmp (in_plugin->name, "avs") ||
+        !strcmp (in_plugin->name, "aiff") ||
+        !strcmp (in_plugin->name, "4xm") ||
+        !strcmp (in_plugin->name, "yuv4mpegpipe"))
       rank = GST_RANK_MARGINAL;
     else
       rank = GST_RANK_NONE;
@@ -1506,14 +1574,9 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     params->videosrccaps = videosrccaps;
     params->audiosrccaps = audiosrccaps;
 
-    g_hash_table_insert (global_plugins,
-        GINT_TO_POINTER (0), (gpointer) params);
-
     /* create the type now */
     type = g_type_register_static (GST_TYPE_ELEMENT, type_name, &typeinfo, 0);
-
-    g_hash_table_insert (global_plugins,
-        GINT_TO_POINTER (type), (gpointer) params);
+    g_type_set_qdata (type, GST_FFDEMUX_PARAMS_QDATA, (gpointer) params);
 
     if (in_plugin->extensions)
       extensions = g_strsplit (in_plugin->extensions, " ", 0);
@@ -1540,7 +1603,6 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     g_free (name);
     in_plugin = in_plugin->next;
   }
-  g_hash_table_remove (global_plugins, GINT_TO_POINTER (0));
 
   return TRUE;
 }
